// SpineBushingTest.cpp
// Author: Adila Faruk
// Based on original CoupledBushingForce class by Ajay Seth
// and SimTk LinearBushing class by Michael Sherman
/*
 * Copyright (c)  2010, Stanford University. All rights reserved. 
* Use of the OpenSim software in source form is permitted provided that the following
* conditions are met:
* 	1. The software is used only for non-commercial research and education. It may not
*     be used in relation to any commercial activity.
* 	2. The software is not distributed or redistributed.  Software distribution is allowed 
*     only through https://simtk.org/home/opensim.
* 	3. Use of the OpenSim software or derivatives must be acknowledged in all publications,
*      presentations, or documents describing work in which OpenSim or derivatives are used.
* 	4. Credits to developers may not be removed from executables
*     created from modifications of the source.
* 	5. Modifications of source code must retain the above copyright notice, this list of
*     conditions and the following disclaimer. 
* 
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
*  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
*  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
*  SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
*  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
*  OR BUSINESS INTERRUPTION) OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
*  WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

//=============================================================================
// INCLUDES
//=============================================================================
#include <OpenSim/Simulation/Model/BodySet.h>
#include <OpenSim/Simulation/Model/Model.h>

#include "SpineBushingTest.h"

//=============================================================================
// STATICS
//=============================================================================
using namespace std;
using namespace SimTK;
using namespace OpenSim;

//=============================================================================
// CONSTRUCTOR(S) AND DESTRUCTOR
//=============================================================================
//_____________________________________________________________________________
/**
 * Destructor.
 */
SpineBushingTest::~SpineBushingTest()
{
}
//_____________________________________________________________________________
/**
 * Default constructor.
 */
SpineBushingTest::SpineBushingTest() : Force(),
	_body1Name(_body1NameProp.getValueStr()),
	_body2Name(_body2NameProp.getValueStr()),
	_printForceAtCoM(_printForceAtCoMProp.getValueBool()),
	_printForceInBodyFrames(_printForceInBodyFramesProp.getValueBool())
{
	setNull();
}

/* Convenience constructor */
SpineBushingTest::SpineBushingTest( std::string body1Name, std::string body2Name, 
									bool printForceAtCoM, bool printForceInBodyFrames,
									SimTK::Mat66 stiffnessMat):	Force(),
	_body1Name(_body1NameProp.getValueStr()),
	_body2Name(_body2NameProp.getValueStr()),
	_printForceAtCoM(_printForceAtCoMProp.getValueBool()),
	_printForceInBodyFrames(_printForceInBodyFramesProp.getValueBool())
{
	setNull();
	_body1Name = body1Name;
	_body2Name = body2Name;
	_printForceAtCoM = printForceAtCoM;
	_printForceInBodyFrames = printForceInBodyFrames;
	_stiffnessMatrix = stiffnessMat;
	updatePropertiesFromMatrices();
}

//// flag for reference frame
//	void setPointsAreGlobal(bool aBool) {_pointsAreGlobal = aBool; };
//	bool getPointsAreGlobal() {return _pointsAreGlobal; };

//_____________________________________________________________________________
/**
 * Copy constructor.
 *
 * @param aForce SpineBushingTest to be copied.
 */
SpineBushingTest::SpineBushingTest(const SpineBushingTest &aForce) :
   Force(aForce),
	_body1Name(_body1NameProp.getValueStr()),
	_body2Name(_body2NameProp.getValueStr()),
	_printForceAtCoM(_printForceAtCoMProp.getValueBool()),
	_printForceInBodyFrames(_printForceInBodyFramesProp.getValueBool())
{
	setNull();
	copyData(aForce);
}

//=============================================================================
// CONSTRUCTION
//=============================================================================
//_____________________________________________________________________________
/**
 * Copy this body and return a pointer to the copy.
 * The copy constructor for this class is used.
 *
 * @return Pointer to a copy of this OpenSim::Body.
 */
Object* SpineBushingTest::copy() const
{
	SpineBushingTest *Force = new SpineBushingTest(*this);
	return(Force);
}
//_____________________________________________________________________________
/**
 * Copy data members from one SpineBushingTest to another.
 *
 * @param aForce SpineBushingTest to be copied.
 */
void SpineBushingTest::copyData(const SpineBushingTest &aForce)
{
	_body1Name = aForce._body1Name;
	_body2Name = aForce._body2Name;
	_printForceAtCoM = aForce._printForceAtCoM;
	_printForceInBodyFrames = aForce._printForceInBodyFrames;

	_stiffnessMatrix = aForce._stiffnessMatrix;
	_stiffnessMatrixRow1Prop = aForce._stiffnessMatrixRow1Prop;
	_stiffnessMatrixRow2Prop = aForce._stiffnessMatrixRow2Prop;
	_stiffnessMatrixRow3Prop = aForce._stiffnessMatrixRow3Prop;
	_stiffnessMatrixRow4Prop = aForce._stiffnessMatrixRow4Prop;
	_stiffnessMatrixRow5Prop = aForce._stiffnessMatrixRow5Prop;
	_stiffnessMatrixRow6Prop = aForce._stiffnessMatrixRow6Prop;
}

//_____________________________________________________________________________
/**
 * Set the data members of this SpineBushingTest to their null values.
 */
void SpineBushingTest::setNull()
{
	setAuthors("Ryan Byrne");
	_b1 = NULL;
	_b2 = NULL;
	setupProperties();
}

//_____________________________________________________________________________
/**
 * Connect properties to local pointers.
 */
void SpineBushingTest::setupProperties()
{	
	// Set the stiffness matrix elements to 0.0 initially
	SimTK::Vec6 defaultVal(0.0, 0.0, 0.0, 0.0, 0.0, 0.0); 

	// Body 1 name
	_body1NameProp.setName("body_1");
	_propertySet.append(&_body1NameProp);

	// Body 2 name
	_body2NameProp.setName("body_2");
	_propertySet.append(&_body2NameProp);

	//// joint connecting Body 2 to Body 1
	//_jointb1b2NameProp.setName("name_of_connecting_joint");
	//_propertySet.append(&_jointb1b2NameProp);

	// Print forces exerted at the body's CoM
	_printForceAtCoMProp.setName("reportForcesAtCoM");
	_printForceAtCoMProp.setComment("True: Prints forces/torques at the CoM. Else, prints F/M at body origins");
	_printForceAtCoMProp.setValue(true);
	_propertySet.append(&_printForceAtCoMProp);

	// Print forces in the body frames
	_printForceInBodyFramesProp.setName("reportForcesInBodyFrames");
	_printForceInBodyFramesProp.setComment("True: Prints forces/torques in the body frames. Else, prints in ground frame");
	_printForceInBodyFramesProp.setValue(true);
	_propertySet.append(&_printForceInBodyFramesProp);

	// Stiffness matrix elements
	_stiffnessMatrixRow1Prop.setName("stiffness_row1");
	_stiffnessMatrixRow1Prop.setValue(defaultVal);
	_propertySet.append(&_stiffnessMatrixRow1Prop );
	_stiffnessMatrixRow2Prop.setName("stiffness_row2");
	_stiffnessMatrixRow2Prop.setValue(defaultVal);
	_propertySet.append(&_stiffnessMatrixRow2Prop );
	_stiffnessMatrixRow3Prop.setName("stiffness_row3");
	_stiffnessMatrixRow3Prop.setValue(defaultVal);
	_propertySet.append(&_stiffnessMatrixRow3Prop );
	_stiffnessMatrixRow4Prop.setName("stiffness_row4");
	_stiffnessMatrixRow4Prop.setValue(defaultVal);
	_propertySet.append(&_stiffnessMatrixRow4Prop );
	_stiffnessMatrixRow5Prop.setName("stiffness_row5");
	_stiffnessMatrixRow5Prop.setValue(defaultVal);
	_propertySet.append(&_stiffnessMatrixRow5Prop );
	_stiffnessMatrixRow6Prop.setName("stiffness_row6");
	_stiffnessMatrixRow6Prop.setValue(defaultVal);
	_propertySet.append(&_stiffnessMatrixRow6Prop );
}

//_____________________________________________________________________________
/**
 * Perform some set up functions that happen after the
 * object has been deserialized or copied.
 *
 * @param aModel OpenSim model containing this SpineBushingTest.
 */
void SpineBushingTest::connectToModel(Model& aModel)
{
	string errorMessage;

	constructMatricesFromProperties();
	// Base class
	Super::connectToModel(aModel);

	// Look up the two bodies being connected by bushing by name in the
	// model and might as well keep a pointer to them
	if (!aModel.updBodySet().contains(_body1Name)) {
		errorMessage = "Invalid bushing body1 (" + _body1Name + ") specified in Force " + getName();
		throw (Exception(errorMessage.c_str()));
	}
	if (!aModel.updBodySet().contains(_body2Name)) {
		errorMessage = "Invalid bushing body2 (" + _body2Name + ") specified in Force " + getName();
		throw (Exception(errorMessage.c_str()));
	}
	
	// Check that body2's joint orientation and orientationInParent are both Vec3(0);
	// (Would be more efficient if I created a subfunction...)
	Body &body2 = aModel.updBodySet().get(_body2Name);
	SimTK::Vec3 orientationVecInBody1(0); body2.getJoint().getOrientationInParent(orientationVecInBody1);
	SimTK::Vec3 orientationVecInBody2(0); body2.getJoint().getOrientation(orientationVecInBody2);
	for (int j=0; j < 3; j++){
            if (orientationVecInBody1(j) != 0) {
				errorMessage = "Joint <orientation_in_parent> of Body " + _body2Name + " needs to be (0 0 0)" ;
				throw (Exception(errorMessage.c_str()));
			}
	}
	for (int j=0; j < 3; j++){
			if (orientationVecInBody2(j)!=0) {
				errorMessage = "Joint <orientation> of Body " + _body2Name + " needs to be (0 0 0)" ;
				throw (Exception(errorMessage.c_str()));
			}
    }

	// Check if the stiffness matrix is symmetric and issue a warning if not.
	bool outputWarning(false);
	for (int i=0; i < 6; i++){
		for (int j=0; j < 6; j++){
			if(_stiffnessMatrix[i][j]!=_stiffnessMatrix[j][i]){
				errorMessage = getName() + " has an asymmetric stiffness Matrix." ;
				outputWarning = true;		break;
			}
		}
	}
	if(outputWarning){
		std::cout << errorMessage << std::endl;
	}

}

void SpineBushingTest::addToSystem(SimTK::MultibodySystem& system) const
{
	Super::addToSystem(system);
	
	Body &body1 = _model->updBodySet().get(_body1Name);
	Body &body2 = _model->updBodySet().get(_body2Name);


	// Beyond the const Component get access to underlying SimTK elements
	SpineBushingTest* mutableThis = const_cast<SpineBushingTest *>(this);

	// Get underlying mobilized bodies
	mutableThis->_b1 = &_model->updMatterSubsystem().getMobilizedBody(body1.getIndex());
	mutableThis->_b2 = &_model->updMatterSubsystem().getMobilizedBody(body2.getIndex());
	
	// Set the default value of the forces to zero
	_f = Vec6(0);
	_F_GCM1  = SpatialVec(0); _F_GCM2  = SpatialVec(0); _F_B1CM1 = SpatialVec(0); _F_B2CM2 = SpatialVec(0);
	_F_GB1   = SpatialVec(0); _F_GB2   = SpatialVec(0); _F_B1    = SpatialVec(0); _F_B2    = SpatialVec(0);

}

//=============================================================================
// OPERATORS
//=============================================================================
//_____________________________________________________________________________
/**
 * Assignment operator.
 *
 * @return Reference to this object.
 */
SpineBushingTest& SpineBushingTest::operator=(const SpineBushingTest &aForce)
{
	Force::operator=(aForce);
	copyData(aForce);
	return(*this);
}

//=============================================================================
// COMPUTATION
//=============================================================================

/**
 * Compute the force contribution to the system and add in to appropriate
 * bodyForce and/or system generalizedForce.
 * SpineBushingTest implementation based on the SimTK::Force::LinearBushing
 * developed and implemented by Michael Sherman.
 */
 
void SpineBushingTest::computeForce(const SimTK::State& s, 
							  SimTK::Vector_<SimTK::SpatialVec>& bodyForces, 
							  SimTK::Vector& generalizedForces) const
{ 
	// RELATIVE MOTION BETWEEN THE BODIES ---------
	// --------------------------------------------

	// Transforms giving placement of body1 and body2's frames (B1 & B2) in
	// in ground
	const Transform& X_GB1 = _b1->getBodyTransform(s);
    const Transform& X_GB2 = _b2->getBodyTransform(s);
	
   // Transforms giving the frames placement in their respective bodies.
   //**Set to the Frames of the Joint connecting B2 to B1** 
   const Transform& X_B1F =_b2->getInboardFrame(s); // placement in B1
   const Transform& X_B2M =_b2->getOutboardFrame(s); // placement in B2

   // Express the bushing frame locations in Ground
   const Transform X_GF = X_GB1 * X_B1F;   
   const Transform X_GM = X_GB2 * X_B2M;  
   // Transform specifying bushing frame M in F. 
   // ** Set to expressing child joint frame in parent joint frame.
   const Transform& X_FM = _b2->getMobilizerTransform(s);
   
   // Create a 6x1 vector consisting of the relative Euler x-y-z 
   // rotational and translational deviations, measured in F's frame
	Vec6 dq(0);
	dq.updSubVec<3>(0) = X_FM.R().convertRotationToBodyFixedXYZ(); 
	dq.updSubVec<3>(3) = X_FM.p();
	 
	// Transformations defining the CoM relative to its body origin
	const Vec3&      p_B1CM   = _b1->getBodyMassCenterStation(s); //in B1 frame
	const Vec3&      p_B2CM   = _b2->getBodyMassCenterStation(s); //in B2 frame
	
	// Re-express local vectors in the Ground frame.
    // Re-express local vectors in the Ground frame.
    const Vec3 p_B1CM_G = X_GB1.R() * p_B1CM;
	const Vec3 p_B2CM_G = X_GB2.R() * p_B2CM;

	// --------------------------------------------
	// FORCES -------------------------------------  

	// Forces and Moments from the Stiffness Matrix acting on Child body. 
	// Forces act at CoMs of B1 and B2, given in F's basis, 
	// Moments act at CoM of B2, given in relative dual Euler basis.
	// Thsee basis vectors were chosen to reflect how the elements of
	// the stiffness matrix are determined experimentally
	_f = -_stiffnessMatrix*dq;
	
	// Assign them to Vec3s
    const Vec3& mCM2_q = _f.getSubVec<3>(0); // Moments, in rel. dual Euler basis, acting at CoM of B2
	const Vec3& fCM2_F = _f.getSubVec<3>(3); // Forces on B2, acting at CoM of B2, expressed in M
	
	// Transform the moments from the Dual Euler basis into the basis
	// vectors attached to frame M. That is, ~N gives the moment in the M frame.
	// So we multiply by X_FM.R() to get it into the F frame
    const Mat33 N_FM  = Rotation::calcNForBodyXYZInBodyFrame(dq.getSubVec<3>(0)); 
	const Vec3  mCM2_F = X_FM.R() * ~N_FM * mCM2_q; // moment acting on body 2, exp. in F
	
	// Transform forces/moments acting at the CoM into the ground frame
	// the negative of this goes on CM1. 
	// All acting in the F-frame
    _F_GCM2 = SpatialVec(  X_GF.R() * mCM2_F,  X_GF.R()*fCM2_F);
    _F_GCM1 = - _F_GCM2;

	// Forces acting at body origins, expressed in ground frame.
	// Involves adding a "r (cross) F" term to the moments.
	_F_GB2 = SpatialVec(_F_GCM2[0] + p_B2CM_G % _F_GCM2[1], _F_GCM2[1]);
	_F_GB1 = SpatialVec(_F_GCM1[0] + p_B1CM_G % _F_GCM1[1], _F_GCM1[1]);

	// Same as above, but with everything in the frames attached to the bodies
	_F_B2CM2 =   SpatialVec(	~X_FM.R()* mCM2_F, ~X_FM.R()*fCM2_F); // at CoM of B2
    _F_B1CM1 = - SpatialVec(			   mCM2_F,           fCM2_F); // at CoM of B1
	_F_B2 = SpatialVec(~X_GB2.R()*_F_GB2[0],  ~X_GB2.R()*_F_GB2[1]);  // in B2 frame
	_F_B1 = SpatialVec(~X_GB1.R()*_F_GB1[0],  ~X_GB1.R()*_F_GB1[1] ); // in B1 frame
	
    // Apply (add-in) the body forces to the system set of body forces
    bodyForces[_b2->getMobilizedBodyIndex()] += _F_GB2;
    bodyForces[_b1->getMobilizedBodyIndex()] += _F_GB1;
}

/** Potential energy stored in the bushing is purely a function of the deflection
    from the rest position and the stiffness.
	As we don't use it here, I set it to 0 so we don't have to create/use a "computeDeflection" function. */
double SpineBushingTest::computePotentialEnergy(const SimTK::State& s) const
{
	double U = 0;
	return U;
}


//=============================================================================
// Reporting
//=============================================================================
/** 
 * Provide names of the quantities (column labels) of the force value(s) reported
 * 
 */
OpenSim::Array<std::string> SpineBushingTest::getRecordLabels() const 
{
	OpenSim::Array<std::string> labels("");

	labels.append(getName()+"."+_body1Name+".FX");
	labels.append(getName()+"."+_body1Name+".FY");
	labels.append(getName()+"."+_body1Name+".FZ");
	labels.append(getName()+"."+_body1Name+".MX");
	labels.append(getName()+"."+_body1Name+".MY");
	labels.append(getName()+"."+_body1Name+".MZ");
	labels.append(getName()+"."+_body2Name+".FX");
	labels.append(getName()+"."+_body2Name+".FY");
	labels.append(getName()+"."+_body2Name+".FZ");
	labels.append(getName()+"."+_body2Name+".MX");
	labels.append(getName()+"."+_body2Name+".MY");
	labels.append(getName()+"."+_body2Name+".MZ");

	return labels;
}
/**
 * Provide the value(s) to be reported that correspond to the labels
 */
OpenSim::Array<double> SpineBushingTest::getRecordValues(const SimTK::State& state) const 
{
	OpenSim::Array<double> values(1);
	
	SimTK::Vec3 forces_b1(0); SimTK::Vec3 torques_b1(0);
	SimTK::Vec3 forces_b2(0); SimTK::Vec3 torques_b2(0);
	// Now decide what to report.
	if(_printForceInBodyFrames){		
		if(_printForceAtCoM){
			// Print the Forces acting at the body CoMs, in the Body-Fixed Frames
			forces_b1 = _F_B1CM1[1]; torques_b1 = _F_B1CM1[0]; 
			forces_b2 = _F_B2CM2[1]; torques_b2 = _F_B2CM2[0];
		} else {
			// Print the forces acting at at the body Origins, in the 
			// in the Body-Fixed Frames
			forces_b1 = _F_B1[1]; torques_b1 = _F_B1[0]; 
			forces_b2 = _F_B2[1]; torques_b2 = _F_B2[0];
		}
	}
	else{
		if(_printForceAtCoM){
			// Forces at body's CoMs, in Ground frame
			forces_b1 = _F_GCM1[1]; torques_b1 = _F_GCM1[0]; 
			forces_b2 = _F_GCM2[1]; torques_b2 = _F_GCM2[0];
		} else {
			// Forces at body's origins, in Ground frame
			forces_b1 = _F_GB1[1]; torques_b1 = _F_GB1[0]; 
			forces_b2 = _F_GB2[1]; torques_b2 = _F_GB2[0];
		}
	}

	values.append(3, &forces_b1[0]);
	values.append(3, &torques_b1[0]);
	values.append(3, &forces_b2[0]);
	values.append(3, &torques_b2[0]);
	return values;
}

/* UTILITIES */
void SpineBushingTest::constructMatricesFromProperties()
{
	_stiffnessMatrix = Mat66(~_stiffnessMatrixRow1Prop.getValueDblVec(),
							 ~_stiffnessMatrixRow2Prop.getValueDblVec(),
							 ~_stiffnessMatrixRow3Prop.getValueDblVec(),
							 ~_stiffnessMatrixRow4Prop.getValueDblVec(),
							 ~_stiffnessMatrixRow5Prop.getValueDblVec(),
							 ~_stiffnessMatrixRow6Prop.getValueDblVec());
}

void SpineBushingTest::updatePropertiesFromMatrices()
{
	_stiffnessMatrixRow1Prop.setValue(_stiffnessMatrix(0));
	_stiffnessMatrixRow2Prop.setValue(_stiffnessMatrix(1));
	_stiffnessMatrixRow3Prop.setValue(_stiffnessMatrix(2));
	_stiffnessMatrixRow4Prop.setValue(_stiffnessMatrix(3));
	_stiffnessMatrixRow5Prop.setValue(_stiffnessMatrix(4));
	_stiffnessMatrixRow6Prop.setValue(_stiffnessMatrix(5));
}