#ifndef __SpineBushing_h__
#define __SpineBushing_h__

/*
* Copyright (c) 2011, Stanford University. All rights reserved. 
* Use of the OpenSim software in source form is permitted provided that the following
* conditions are met:
* 	1. The software is used only for non-commercial research and education. It may not
*     be used in relation to any commercial activity.
* 	2. The software is not distributed or redistributed.  Software distribution is allowed 
*     only through https://simtk.org/home/opensim.
* 	3. Use of the OpenSim software or derivatives must be acknowledged in all publications,
*      presentations, or documents describing work in which OpenSim or derivatives are used.
* 	4. Credits to developers may not be removed from executables
*     created from modifications of the source.
* 	5. Modifications of source code must retain the above copyright notice, this list of
*     conditions and the following disclaimer. 
* 
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
*  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
*  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
*  SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
*  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
*  OR BUSINESS INTERRUPTION) OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
*  WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE

 * @author Adila Faruk
 * Based on previous work by Michael Sherman and Ajay Seth.
 * Further details are given in the "SpineBushingCode_ReadMe" document
 * and the cited publication.
 * 
 * @version 1.0>
 */


// INCLUDE
#include <string>
#include "osimPluginDLL.h"
#include <OpenSim/OpenSim.h>
#include <OpenSim/Common/Object.h>
#include <OpenSim/Common/PropertyStr.h>
#include <OpenSim/Common/PropertyDblVec.h>
#include <OpenSim/Simulation/Model/Force.h>

namespace OpenSim {

//=================================================================================
//=================================================================================
/*
 * A class implementing a SpineBushingForce.
 * Details on this function and its correct use are given in 
 ** "On the Modeling of the Intervertebral Disc in Multibody Models
 **  of the Spine", submitted for publication.
 * 
 
 // A typical Coupled-Bushing-Force function -------------------------
 * A Couple-Bushing-Force is the force proportional to the deviation of two frames. 
 * That is, the coupled bushing force is a *generalized* force given by 
 * [K]*dq + [D]*dqdot. Here,
 * [K]    is the spatial 6dof stiffness matrix between the two frames 
 * dq     is the deflection in body spatial coordinates with the first three 
 *        components corresponding to the body2 fixed x-y-z Euler angles and the
 *	      next 3 given by the translation of the bushing frame on body2, M with respect to
 * 	      the bushing frame on body1, F. 
 * [D]    is the spatial 6dof damping matrix opposing the velocity between the frames given 
 * dqdot  is the relative spatial velocity of the two frames with the first three 
 *        components corresponding to the rate of change of three (body2 fixed) x-y-z 
 *        Euler angles and the next 3 equal to the translational velocity of the bushing 
 * 	      frame on body2, M, with respect to the bushing frame on body1, F. 
 *  
 * The stiffness and damping matrices can be written as
 *		 K = [K_1 K_3]				D = [D_1 D_3]
 *           [K_4 K_2],					[D_4 D_2],
 * where the K_i and D_i are all 3x3 stiffness matrices.
 * K_1  relates changes in the rotational coordinates to the generalized moments.
 * K_2  relates changes in the translational coordinates to the generalized forces.
 * K_3  relates changes in the translational coordinates to the generalized moments.
 * K_4  relates changes in the rotational coordinates to the generalized forces.
 * and
 * D_1  relates the time rate of change of the Euler angles to the generalized moments.
 * D_2  relates the time rate of change of the translational coordinates 
 *      to the generalized forces.
 * D_3  relates the time rate of change of the translational coordinates 
 *      to the generalized moments.
 * D_4  relates the time rate of change of the Euler angles to the generalized forces.
 *
 * Appropriate transformations are then performed to express this force 
 * (acting on the M frame in B2 and the F frame in B1) in the Ground frame.
 * 
 *
 // The SpineBushingForce ----------------------
 * In the remainder, we focus **only on the stiffness matrix**. The algorithm for including
 * damping is similar, but involves translational and angular velocity terms. 
 *
 *
 * The SpineBushingForce differs from the Coupled-Bushing-Force in a number of ways:
 * (1) The forces and moments act at the *same points used to compute K experimentally* 
 *     We assume it's the center of mass (COM) here, and that the vertebral Centers of
 *     mass are situated at the vertebral centers of geometry. 
 *     If you are using a stiffness matrix that's computed using different points of 
 *     application, then you'll have to edit this. 
 *
 * (2) We use the *CHANGE IN RELATIVE MOTION* between the CoMs from the neutral posture
 *	   to compute the generalized displacement vector. That is, at neutral, dq = 0 
 *     to reduce on the number of inputs needed, we just use the relative displacement/
 *     orientation between the joint frames since the joint frames are initially 
 *     coincident with each other.
 ** but, CAUTION: 
 * Make sure that the *orientation* of the joint frames
 * (i.e. <orientation> and <orientation_in_parent>) are all zero relative
 * to the parent and child bodies or you'll mess things up. If you are forced to set them
 * to non-zero values (by knife-point) then you will have to edit the force and moment
 * computations and also take in more inputs.
 *
 * (3) Going backwards from K to a potential energy function isn't as simple as writing
 *     U = .5 * (q^T) * K * q. 
 *     There's an extra term that comes from the Taylor series expansion of the
 *     moving basis.
 *     We show this in the cited journal article, and in my (very long!) dissertation. 
 *     For this reason, the computePotentialEnergy function returns a PE = 0;
 *
 */
class OSIMPLUGIN_API SpineBushingTest : public Force {
OpenSim_DECLARE_CONCRETE_OBJECT(SpineBushingTest, Force);

/* This is for if you want to run it in a .cpp executable.
 You'll need to build a different solution file though using a separate 
 CMakeLists.txt */
//class SpineBushingTest : public Force  


//=============================================================================
// DATA
//=============================================================================
protected:
	/** Specify first of two bodies held together by the bushing. */
	PropertyStr _body1NameProp;
	std::string& _body1Name;

	/** Specify second of two bodies held by the bushing force. */
	PropertyStr _body2NameProp;
	std::string& _body2Name;

	/** Specify if we want to print the forces/torque exerted on the 
		body origins, or forces/torques exerted on the body Centers of Mass.
		Default is the latter (torques at CoMs) */
	PropertyBool _printForceAtCoMProp;
	bool& _printForceAtCoM;

	/** Specify if we want to print the forces/torque exerted at the CoMs
	    in the body-fixed frame of reference. Default is true, but this is
		only called *iff* the forces/moments are exerted at the CoMs of
		the respective bodies.*/
	PropertyBool _printForceInBodyFramesProp;
	bool& _printForceInBodyFrames;

	/** Stiffness of the bushing related to Euler XYZ body-fixed angular and
	    translationals deviations that express frame2 in frame1. Force is zero
	    when bushing frames are coincident and aligned with the frames of body1
	    and body2.  */
	PropertyDblVec6 _stiffnessMatrixRow1Prop;
	PropertyDblVec6 _stiffnessMatrixRow2Prop;
	PropertyDblVec6 _stiffnessMatrixRow3Prop;
	PropertyDblVec6 _stiffnessMatrixRow4Prop;
	PropertyDblVec6 _stiffnessMatrixRow5Prop;
	PropertyDblVec6 _stiffnessMatrixRow6Prop;
	SimTK::Mat66 _stiffnessMatrix;

private:
	// underlying SimTK system elements

	// the mobilized bodies involved
	const SimTK::MobilizedBody *_b1;
	const SimTK::MobilizedBody *_b2;
	

	/**Forces at various points and in different frames. Might as well keep track of
	them so we can use it in the ForceReporter without having to recompute everything. 
	Set to mutable so we can update these in the  OpenSim::ComputeForce() const function. */
	 
	/* generalized force vector from the Stiffness Matrix acting on _b2 body. 
	Forces act at CoMs of _b2, and are given in F's basis, 
	Moments act at CoM of _b2, given in relative dual Euler basis.
	Thsee basis vectors were chosen to reflect how the elements of the stiffness matrix 
	are determined experimentally. 
	Note: it's Moments first, then Forces*/
    mutable SimTK::Vec6				_f;   

	/* Forces/Moments at the CoMs, expressed in the ground frame. These
	are just the relevant components of _f, transformed into the ground frame */
	mutable SimTK::SpatialVec _F_GCM1; // at CoM of body 1, 
	mutable SimTK::SpatialVec _F_GCM2;  // at CoM of body 2, 

	/* Forces/Moments at the CoMs, expressed in the body frames. These
	are just the relevant components of _f, transformed into the frames attached to each body */
	mutable SimTK::SpatialVec _F_B1CM1; // at CoM of body 1, 
	mutable SimTK::SpatialVec _F_B2CM2;  // at CoM of body 2,


	/* Forces/Moments at the body frames, expressed in the ground frame. The force
	component is the same as _F_GCM, but the moment has an additional r (cross) F term
	to account for the shift. */
    mutable SimTK::SpatialVec _F_GB1; // at B1 body frame
    mutable SimTK::SpatialVec _F_GB2;    // at B2 body frame

	/* Forces/Moments at the body frames, expressed in the body fixed frame. The force
	component is the same as _F_GCM, but the moment has an addition r (cross) F term
	to account for the shift. */
    mutable SimTK::SpatialVec _F_B1; // at B1 body frame
    mutable SimTK::SpatialVec _F_B2;    // at B2 body frame
	 

	

//=============================================================================
// METHODS
//=============================================================================
public:
	// CONSTRUCTION
	SpineBushingTest();
	SpineBushingTest( std::string body1Name, std::string body2Name, 
					  bool printForceAtCoM, bool printForceInBodyFrames,
				  SimTK::Mat66 stiffnessMat);
	SpineBushingTest(const SpineBushingTest &aForce);
	virtual ~SpineBushingTest();
	virtual Object* copy() const;
	SpineBushingTest& operator=(const SpineBushingTest &aForce);
	void copyData(const SpineBushingTest &aForce);

//=============================================================================
// GET and SET
//=============================================================================
//_____________________________________________________________________________
	/**
	 * The following methods set attributes of the Spine Bushing Force */
	void setBody1ByName(std::string aBodyName)
	{	_body1Name = aBodyName;		}

	void setBody2ByName(std::string aBodyName)
	{	_body2Name = aBodyName;		}

	void setPrintForceAtCoM(bool aPrintForceAtCoM)
	{	_printForceAtCoM = aPrintForceAtCoM;	 }

	void setPrintForceInBodyFrames(bool aPrintForceInBodyFrames)
	{	_printForceInBodyFrames = aPrintForceInBodyFrames;		 }

	/**
	 * Get function 
	 */
	SimTK::Vec6 getf(const SimTK::State& s) const    { return _f ;   };
	SimTK::SpatialVec getF_GCM1(const SimTK::State& s) const  { return _F_GCM1 ; };
	SimTK::SpatialVec getF_GCM2(const SimTK::State& s) const  { return _F_GCM2 ; };
	SimTK::SpatialVec getF_B1CM1(const SimTK::State& s) const { return _F_GCM1 ; };
	SimTK::SpatialVec getF_B2CM2(const SimTK::State& s) const { return _F_GCM2 ; };
	SimTK::SpatialVec getF_GB1(const SimTK::State& s) const   { return _F_GB1 ;  };
	SimTK::SpatialVec getF_GB2(const SimTK::State& s) const   { return _F_GB2 ;  };
	SimTK::SpatialVec getF_B1(const SimTK::State& s) const    { return _F_B1 ;   };
	SimTK::SpatialVec getF_B2(const SimTK::State& s) const    { return _F_B2 ;   };
	

	//--------------------------------------------------------------------------
	// COMPUTATION
	//--------------------------------------------------------------------------
	
	/** Compute the bushing force contribution to the system and add it in as a
	  * bodyForce. The Spine Bushing Force is the force given by -[K]*dq where,
	  * [K]    is the spatial 6dof stiffness matrix corresponding to the
	  *		   joint, computed using forces and moments exerted at the adjacent
	  *        bodies' CoMs.
	  * dq     is the relative motion in body spatial coordinates. The first three 
	  *        components correspond to the body2 fixed x-y-z Euler angles relative 
	  * 	   to body1; the next 3 are given by the *relative translation* body2 
	  *		   with respect to body1, from the equilibrium position.
	  *		   At neutral, dq = 0.
	  *
	  */
	virtual void computeForce(const SimTK::State& s, 
							  SimTK::Vector_<SimTK::SpatialVec>& bodyForces, 
							  SimTK::Vector& generalizedForces) const;


    /** Determine the potential energy stored in the bushing, defined as 
    * U = 0.5* (~dq*[H]*dq ), where H is the Hessian of U and dq is defined above. 
    * It is computed here as U = 0.5* (~dq*[K]*dq) under the assumption that the user
    * only trusts the results when K and H are equal to each other.*
    ** In General, This Is Not True!** 
    * It occurs in certain special situations though. E.g., when K is obtained from 
    * pure relative translational motion or relative rotation about a single axis (with 
	* the moments computed in the dual Euler basis. For the purposes of this SpineBushing,
	* U = 0 even though there are situations when this is not right.  */
 	virtual double computePotentialEnergy(const SimTK::State& s) const;

	//-----------------------------------------------------------------------------
	// Reporting
	//-----------------------------------------------------------------------------
	/** 
	 * Provide name(s) of the quantities (column labels) of the force value(s) to be reported
	 */
	virtual OpenSim::Array<std::string> getRecordLabels() const ;
	/**
	*  Provide the value(s) to be reported that correspond to the labels
	*/
	virtual OpenSim::Array<double> getRecordValues(const SimTK::State& state) const ;

protected:
	void connectToModel(Model& aModel) OVERRIDE_11;
	/**
	 * Create a SimTK::Force::LinearBushing which implements this SpineBushingTest.
	 */
	void addToSystem(SimTK::MultibodySystem& system) const OVERRIDE_11;


private:
	void setNull();
	void setupProperties();
	void constructMatricesFromProperties();
	void updatePropertiesFromMatrices();

//=============================================================================
};	// END of class SpineBushingTest
//=============================================================================
//=============================================================================

} // end of namespace OpenSim

#endif // __SpineBushing_h__


